//------------------------------------------------------------------------------
// File: Compressor.cpp
//------------------------------------------------------------------------------

/*
 based on xmms-compress v1.1 (2002/07/31)
 (c)2002 trikuare studios (http://trikuare.cx)

 http://trikuare.cx/~magenta/projects/xmms-compress.html
*/

/*
* Strongly modified for ReClock by OGO :)
*
*/

#include <atlbase.h>
#include <streams.h>
#include <initguid.h>

#include "dxutil.h"
#include "compressor.h"

#undef COMPRESSOR_STATS

#define COMPRESSOR_BUCKETS 20		/* How long of a history to store */

Compressor::Compressor(int samplesPerSec, bool anticlip, int target, int gainmax, int gainsmooth, int bucketsTime)
{
	m_peaks = 0;
	Configure(samplesPerSec, anticlip, target, gainmax, gainsmooth, bucketsTime);
}

Compressor::Configure(int samplesPerSec, bool anticlip, int target, int gainmax, int gainsmooth, int bucketsTime)
{
	m_anticlip = anticlip;
	m_target = target;
	m_gainmax = gainmax;
	m_gainsmooth = gainsmooth;
	m_bucketsTime = bucketsTime;

	m_gainCurrent = m_gainTarget = (1 << COMPRESSOR_GAINSHIFT);
	
	m_clip = 0;
	
	/* Allocate peaks */
	SAFE_DELETE_ARRAY(m_peaks);
	m_peaks = new int[COMPRESSOR_BUCKETS];
	for (int i = 0; i < COMPRESSOR_BUCKETS; i++)
		m_peaks[i] = 0;
	m_currentBucketCount = 0;
	m_currentBucketPeak = 1;
	m_samplesPerBucket = (samplesPerSec * m_bucketsTime) / COMPRESSOR_BUCKETS;
	m_pn = -1;
}

Compressor::~Compressor()
{
	SAFE_DELETE_ARRAY(m_peaks);
}

void Compressor::ProcessFlow16(short *input, short *output, int sampleCount)
{
	/* Determine peak's value and position */
	int l_peak = 1;
	int l_pos = 0;
	
	//ATLTRACE( "finding peak (m_pn=%d)\n", m_pn);
	
	short* l_ap = input;
	for (int i = 0; i < sampleCount; i++)
	{
		int l_val = *l_ap++;
		
		if (l_val > l_peak)
		{
			l_peak = l_val;
			l_pos = i;
		} 
		else if (-l_val > l_peak)
		{
			l_peak = -l_val;
			l_pos = i;
		}

		if (l_val > m_currentBucketPeak)
			m_currentBucketPeak = l_val;
		else if (-l_val > m_currentBucketPeak)
			m_currentBucketPeak = -l_val;
		if (++m_currentBucketCount >= m_samplesPerBucket)
		{
			m_pn = (m_pn + 1) % COMPRESSOR_BUCKETS;
			m_peaks[m_pn] = m_currentBucketPeak;
			m_currentBucketPeak = 1;
			m_currentBucketCount = 0;
		}
	}
	
	for (i = 0; i < COMPRESSOR_BUCKETS; i++)
	{
		if (m_peaks[i] > l_peak)
		{
			l_peak = m_peaks[i];
			l_pos = 0;
		}
	}
	
	/* Determine target gain */
	int l_gn = ((1 << COMPRESSOR_GAINSHIFT) * m_target) / l_peak;
	if (l_gn < (1 << COMPRESSOR_GAINSHIFT))
		l_gn = 1 << COMPRESSOR_GAINSHIFT;
	
	m_gainTarget = (m_gainTarget * ((1 << m_gainsmooth) - 1) + l_gn) >> m_gainsmooth;
	
	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/
	
	if (l_gn < m_gainTarget)
		m_gainTarget--;
	else if (l_gn > m_gainTarget)
		m_gainTarget++;
	
	if (m_gainTarget > m_gainmax << COMPRESSOR_GAINSHIFT)
		m_gainTarget = m_gainmax << COMPRESSOR_GAINSHIFT;
	
	/* See if a peak is going to clip */
	l_gn = ((1 << COMPRESSOR_GAINSHIFT) * 32768) / l_peak;
	
	if (l_gn < m_gainTarget)
	{
		m_gainTarget = l_gn;
		
		if (m_anticlip)
			l_pos = 0;
		
	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		l_pos = (m_samplesPerBucket * COMPRESSOR_BUCKETS);
	}
	
	/* Determine gain rate necessary to make target */
	if (l_pos == 0)
		l_pos = 1;
	
	int l_gr = ((m_gainTarget - m_gainCurrent) << 16) / l_pos;
	
	/* Do the shiznit */
	int l_gf = m_gainCurrent << 16;
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( " gain = %2.2f %+.2e ", m_gainCurrent*1.0/(1 << COMPRESSOR_GAINSHIFT), (m_gainTarget - m_gainCurrent)*1.0/(1 << COMPRESSOR_GAINSHIFT));
#endif
	
	l_ap = input;
	for (i = 0; i < sampleCount; i++)
	{
		/* Interpolate the gain */
		m_gainCurrent = l_gf >> 16;
		if (i < l_pos)
			l_gf += l_gr;
		else if (i == l_pos)
			l_gf = m_gainTarget << 16;
		
		/* Amplify */
		int l_sample = ((*l_ap++) * m_gainCurrent) >> COMPRESSOR_GAINSHIFT;
		if (l_sample < -32768)
		{
			m_clip++;
			l_sample = -32768;
		}
		else if (l_sample > 32767)
		{
			m_clip++;
			l_sample = 32767;
		}

		*output++ = (short)l_sample;
	}
	
#ifdef COMPRESSOR_STATS
	ATLTRACE( "clip %d b%-3d\n", m_clip, m_pn);
#endif
}



